//****************************************************************************
//This file contains declarations and operations for derivation trees in TAGs*
//Programmed by NXH (version 3.2)                                            *
//****************************************************************************
#ifndef __DTREE__
#define __DTREE__
#include<iostream.h>
#include<stdlib.h>
#include<string.h>
#include<assert.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
//node structure for derivation tree node
struct _dnode {
		SIZE id;//the elementary tree id
		BYTE addno;//address number
		struct _dnode *children,*sibling;//the addresses should be incresasing
		llist lexlist;//list of lexicon for substitution
	       };
typedef struct _dnode dnode;
//Operations for derivation tree
//***************************
//Make a Node               *
//***************************
dnode *MakeDNode(SIZE id, BYTE addno)
{
  dnode *t;
  t=new dnode;
  assert(t!=NULL);
  t->id=id;
  t->addno=addno;
  t->lexlist=BuildLList(etree[id].root);
  t->sibling=NULL;
  t->children=NULL;
  return t;
}
void DeleteSDNode(dnode *t)
{
 DeleteLList(t->lexlist);
 delete t;
}
//*****************************
//Delete a subtree            *
//*****************************
void DeleteDNode(dnode *t)
{
  dnode *q,*p;
  if (t)
   {
    if(t->children==NULL)
      DeleteSDNode(t);
     else
      {
	p=t->children;
	while(p)
	  {
	    q=p->sibling;
	    DeleteDNode(p);
	    p=q;
	  }
	 DeleteSDNode(t);
      }
  }
}
//**********************************
//Copy a subtree                   *
//**********************************
dnode *CopyDNode(dnode *s)
{
  dnode *t,*p,*q;
  if(s)
   {
     t=new dnode;
     assert(t!=NULL);
     t->id=s->id;
     t->addno=s->addno;
     t->lexlist=CopyLList(s->lexlist);
     t->sibling=NULL;
     t->children=NULL;

     q=s->children;
     t->children=CopyDNode(q);
     p=t->children;
     while (q)
      {
	p->sibling=CopyDNode(q->sibling);
	q=q->sibling;
	p=p->sibling;
      }
     return t;
   }
  return NULL;
}
//****************************************
//Compare two trees (structure only)     *
//****************************************
BOOL DNodeCmp(dnode *t,dnode *s)
{
 if((t!=NULL) && (s!=NULL))
  {
  if((t->id==s->id)&&(t->addno==s->addno))
   {
     dnode *p=t->children,*q=s->children;
     BOOL b=TRUE;
     while(b && (p!=NULL) && (q!=NULL))
      {
	b=DNodeCmp(p,q);
	p=p->sibling;
	q=q->sibling;
      }
      if (b==FALSE)
	  return b;
       else
	  if ((p==NULL)&&(q==NULL))
	   return TRUE;
	  else
	   return FALSE;
   }
   else
    return FALSE;
  }
 else
  if((t==NULL) && (s==NULL))
   return TRUE;
    else
     return FALSE;
}
int GetHeightDNode(dnode *t)
{
   if (t->children==NULL)
     return 0;
   else
     {
       unsigned int c,max=0;
       dnode *p=t->children;
       while (p)
	{
	  c=GetHeightDNode(p);
	  if (c>max)
	      max=c;
	  p=p->sibling;
	 }
	return max+1;
      }
}
void DisplayDTree(dnode *r)
{
 dnode *p;
 if (r)
   {
    cout<<(int)r->id<<" address: "<<(int)r->addno;
    p=r->children;
    if(p)
     {
      cout<<"(";
      while(p)
       {
	DisplayDTree(p);
	if(p->sibling)
	  cout<<",";
	p=p->sibling;
       }
      cout<<")";
     }
   }
}
//*********************************************************
//Travel a Tree in Preorder,Inorder and Postoder          *
//*********************************************************
void (*_doitdnode)(dnode *n);
void PreorderDNode(dnode *t)
{
  dnode *p;
   if(t)
    {
     _doitdnode(t);
      p=t->children;
      while (p)
       {
	 PreorderDNode(p);
	 p=p->sibling;
       }
    }
}
//Travel a Tree in Inoder
void InorderDNode(dnode *t)
{
  dnode *p;
  if (t)
   {
     p=t->children;
     InorderDNode(p);
     _doitdnode(t);
     while (p)
     {
      p=p->sibling;
      InorderDNode(p);
     }
   }
}
//Travel a Tree in Postorder
void PostorderDNode(dnode *t)
{
 dnode *p;
 if (t)
  {
    p=t->children;
    while (p)
     {
      PostorderDNode(p);
      p=p->sibling;
     }
    _doitdnode(t);
  }
}
//**********************************
//Travel the tre level by level    *
//**********************************
//We need a queue
struct _queuenode {
		     dnode *id;
		     struct _queuenode *next;
		    };
typedef struct _queuenode queuenode;
struct _queue{
	       queuenode *head;
	       queuenode *tail;
	       int card;
	      };
typedef  struct _queue queue;
void IniQueue(queue &q)
{
 q.head=NULL;
 q.tail=NULL;
 q.card=0;
}
BOOL EmptyQueue(queue q)
{
 if (q.head==NULL)
  return TRUE;
 else
  return FALSE;
}
void AddQueue(queue &q,dnode *t)
{
 queuenode *p;
 p=new queuenode;assert(p!=NULL);
 p->id=t;
 p->next=NULL;
 if(EmptyQueue(q))
  {
   q.head=p;
   q.tail=p;
   q.card=1;
  }
  else
   {
    (q.tail)->next=p;
    q.tail=p;
    q.card++;
   }
}
dnode *PopQueue(queue &q)
{
  queuenode *p;
  dnode *t;
 if(EmptyQueue(q))
 {
  cout<<"Error Pop From Empty Queue";
  exit(1);
 }
 else
  {
   p=q.head;
   q.head=(q.head)->next;
   if(q.head==NULL)//return to empty
    q.tail=NULL;
   q.card--;
   t=p->id;
   delete p;
   return t;
  }
 return NULL;
}
void LevelorderDNode(dnode *t)
{
  queue q;
  dnode *p1,*p2;
  if(t)
  {
   IniQueue(q);
   AddQueue(q,t);
   while (!EmptyQueue(q))
    {
     p1=PopQueue(q);
     _doitdnode(p1);
     p2=p1->children;
     while (p2)
      {
       AddQueue(q,p2);
       p2=p2->sibling;
      }
    }
  }
}
//********************************
//Compute the size of a subtree  *
//********************************
int  _dnodecount;
BOOL _countleave;
void DNodeCount(dnode *t)
{
 _dnodecount++;
 if(!_countleave && (t->children==NULL))//if don't count the leaves
 _dnodecount--;
}
int  GetSizeDNode(dnode *t,BOOL countleave)
{
 _countleave=countleave;
 _dnodecount=0;
 _doitdnode=DNodeCount;
 PreorderDNode(t);
 return _dnodecount;
}
//***************************************
//Get the average branching of a dnode  *
//***************************************
int _dnodebranch;
void DNodeBranchCount(dnode *t)
{
 dnode *p;
 BYTE i;
 if(t)
 {
  if (t->children!=NULL)
     {
      i=0;
      p=t->children;
      while(p)
      {
       i++;
       p=p->sibling;
      }
      _dnodebranch+=i;
     }
 }
}
double GetAVGDNode(dnode *t)
{
 int size;
 size=GetSizeDNode(t,FALSE);
 _dnodebranch=0;
 _doitdnode=DNodeBranchCount;
 PreorderDNode(t);
 return ((double)_dnodebranch)/size;
}
//****************************************************
//Check if dnode p is in subtree stemed from dnode q *
//****************************************************
BOOL _indnode;
dnode *_checkdnode;
void CheckDNode(dnode *s)
{
 if(!_indnode&&s)
  {
   if(s==_checkdnode)
    _indnode=TRUE;
  }
}
BOOL InDNode(dnode *p,dnode *q)
{
 _indnode=FALSE;
 _checkdnode=p;
 _doitdnode=CheckDNode;
 PreorderDNode(q);
 if( _indnode)
  return TRUE;
 else
  return FALSE;
}
int DNodeOrder(dnode *p,dnode *q)
{
 if(InDNode(p,q))
  return 1;//p is in q
 else
  if(InDNode(q,p))
   return -1;//q is in p
 else
  return 0;// different branch
}
//*******************************************************
//Find ith node and its parent (in preorder search)     *
//*******************************************************
dnode *_idnode;
dnode *_idnodeparent;
int   _indexdnode;
void FindDNode(dnode *t,dnode *parent)
{
 if(_indexdnode>0)
  {
   if(t)
    {
     dnode *p;
     _indexdnode--;
     if (_indexdnode==0)
      {
       _idnode=t;
       _idnodeparent=parent;
      }
      else
       {
	p=t->children;
	while(p)
	 {
	  FindDNode(p,t);
	  p=p->sibling;
	 }
       }
    }
  }
}
void LocateDNode(int i,dnode *t,dnode *parent)
{
  _indexdnode=i;
  _idnode=NULL;
  _idnodeparent=NULL;
  FindDNode(t,parent);
}

//*************************************************************
//This function copy a sub node  tree and a list of address   *
//*************************************************************
node *_oldfoot2=NULL;//this variable contains the foot node of the source tree
node *_newfoot2=NULL;//this variable contains the footnode of the newtree
node *CopyNode2(node *s, node *parent,SIZE treeindex,addlist &newadd)
{
  node *t,*p,*q;
  address *v;
  BOOL found;
  slist st;
  if(s)
   {
     t=MakeNode(s->name, s->value);
     if (s==_oldfoot2)
	 _newfoot2=t;
 //check whether s is in the addlist of the tree h
     v=adtable[treeindex].head;
     found=FALSE;
      while(v && !found)
	if(v->addr==s)
	 found=TRUE;
	else
	 v=v->next;
     if (found)
      {
       IniSlist(st);
       AddHeadAddList(newadd,st,t,parent,v->constraint);
      }
     q=s->children;
     t->children=CopyNode2(q,t,treeindex,newadd);
     p=t->children;
     while (q)
      {
	p->sibling=CopyNode2(q->sibling,t,treeindex,newadd);
	q=q->sibling;
	p=p->sibling;
      }
     return t;
   }
  return NULL;
}
stree CopyTree2(SIZE treeid,addlist &newadd)
{
 stree v;
 _oldfoot2=etree[treeid].foot;
 _newfoot2=NULL;
 v.root=CopyNode2(etree[treeid].root,NULL,treeid,newadd);
 v.foot=_newfoot2;
 return v;
}
//*************************************************************
//This function decode a derived tree from a derivation tree  *
//Using Weir Function and substitution                        *
//*************************************************************
stree DecodeDTree(dnode *t,BOOL stochastic)
{
 stree st;
 if (t->children==NULL)
 {
  st=CopyTree(etree[t->id]);//if leaves then just copy

  if(!EmptyLList(t->lexlist))
   if(stochastic)//random substitution
     SubstituteLexicon(st.root);
   else
    DSubstituteLexicon(st.root,t->lexlist);
   return st;
 }
 else
  {
    stree beta,gama;
    address *s;
    addlist newadd;
    BYTE i;
    dnode *p;
    IniAddList(newadd);
    gama=CopyTree2(t->id,newadd);
  //substitue
  if(!EmptyLList(t->lexlist))
   if(stochastic)//random substitution
     SubstituteLexicon(gama.root);
   else
    DSubstituteLexicon(gama.root,t->lexlist);
    p=t->children;
   while (p)
    {
    if(p->id>=0) //if the subtree is not currently switched off (subcode type1);
     { 
      beta=DecodeDTree(p,stochastic);
      //find the address and parent
      s=newadd.head;
      i=0;
      while (i<p->addno)
      {
       s=s->next;
       i++;
      }
      gama= AdjointTree(beta,gama,s->addr,s->parent);
     }
    else
     if(p->id<=-numetree)
//if the subtree is currently switched off but the root adjunction is still active (sub-code of type2)
     {
       dnode *q,*t;
       q=p->children;
       while (q->sibling) q=q->sibling;
       while(q->id<=-numetree)
       {
        t=q->children;
        while(t->sibling) t=t->sibling;
        q=t;
       }
     if(q->id>=0)//only decode if it is not switched off
      { 
        beta=DecodeDTree(q,stochastic);
        s=newadd.head;
        i=0;
       while(i<p->addno)
        {
         s=s->next;
         i++;
        }
       gama=AdjointTree(beta,gama,s->addr,s->parent);   
      }
     } 
     p=p->sibling;
    }
  
   DestroyAddList(newadd);
    return gama;
  }
}
//***********************************************************
//Randomly add one more beta tree in to the derivation tree *
//***********************************************************
int _opendnodeno;
dnode *_opendnode;
BOOL _countorfind;
void CountOpenDNode(dnode *s)
{
 BYTE childrenno;
 dnode *t;
 childrenno=0;//count number of children;
 t=s->children;
 while(t)
  {
   childrenno++;
   t=t->sibling;
  }
 if(adtable[s->id].card>childrenno)
  if (_countorfind)
  _opendnodeno++;
  else
   {
    _opendnodeno--;
    if(_opendnodeno==0)
       _opendnode=s;
   }
}
//***********************************************************************************
//This function is used for counting/finding the number of open adjoining addresses *
//***********************************************************************************
int _openaddno;
dnode *_opendnodeadd;
BYTE _openaddindex;
BOOL _cf,_stop;
void CountOpenAddDNode(dnode *s)
{
 BYTE childrenno;
 dnode *t;
 if(!_stop)
{
 childrenno=0;//count number of children;
 t=s->children;
 while(t)
  {
   childrenno++;
   t=t->sibling;
  }
 if(adtable[s->id].card>childrenno)
  if (_cf)
  _openaddno+=adtable[s->id].card-childrenno;
  else
   {
    _openaddno-=adtable[s->id].card-childrenno;
    if(_openaddno<=0)
	{
		_stop=TRUE; //stop the finding
       _opendnodeadd=s;
       _openaddindex=adtable[s->id].card-childrenno+_openaddno; //the order of open address
	}
   }
 }
 }
//Counting Open Addresses
int CountOpenAdd(dnode *s)
{
	_openaddno=0;
    _cf=TRUE;
	_stop=FALSE;
	_doitdnode=CountOpenAddDNode;
    PreorderDNode(s);
	return _openaddno;
  
}
//locating an open address
dnode *LocateOpenAdd(dnode *s,int index,BYTE &add)
{
   BOOL *check;  
   BYTE i;
   dnode *p;
   _openaddno=index;
    _cf=FALSE;
	_stop=FALSE;
	_opendnodeadd=NULL;
	_doitdnode=CountOpenAddDNode;
    PreorderDNode(s);
   // *n=_opendnodeadd;
    //compute the address number

//marking the occupied addresses
 p=_opendnodeadd->children;
 check=new BOOL[adtable[_opendnodeadd->id].card]; assert(check!=NULL);
 for(i=0;i<adtable[_opendnodeadd->id].card;i++)
  check[i]=FALSE;
  while (p)
  {
   check[p->addno]=TRUE;
   p=p->sibling;
  }
  //picking up the opening address
  i=0;
  while((_openaddindex>0) && (i<adtable[_opendnodeadd->id].card))
  {
    if(!check[i])
     _openaddindex--;
    if(_openaddindex>0)
     i++;
  }
  
  delete [] check;
  add=i;
  return _opendnodeadd;
}
BOOL AddBeta(dnode *s)
{
dnode *p,*t,*q;
BYTE childrenno,addno;
BOOL *check;
address *v;
snode *n;
SIZE j,i;
//First count to see how many node is still open
_opendnodeno=0;
_countorfind=TRUE;//counting
_doitdnode=CountOpenDNode;
 PreorderDNode(s);//result is in _opendnodeno
if (_opendnodeno<=0)
 return FALSE;
//Then choose one open node as random
_opendnodeno=IRandom(1,_opendnodeno);
_opendnode=NULL;
_countorfind=FALSE;
 PreorderDNode(s);//result is in _opendnode
//Find the number of open addresses
 childrenno=0;
 p=_opendnode->children;
 check=new BOOL[adtable[_opendnode->id].card]; assert(check!=NULL);
 for(i=0;i<adtable[_opendnode->id].card;i++)
  check[i]=FALSE;
 while (p)
 {
   check[p->addno]=TRUE;
   childrenno++;
   p=p->sibling;
 }
 childrenno=IRandom(1,adtable[_opendnode->id].card-childrenno);
 i=0;
 while((i<adtable[_opendnode->id].card)&&(childrenno>0))
 {
  if(!check[i])
   childrenno--;
  if(childrenno>0)
  i++;
 }
 delete [] check;
 addno=i;//the index of the open address
//Find a random beta tree in the source list of the address
 i=0;
 v=adtable[_opendnode->id].head;
 while(i<addno)
 {
  v=v->next;
  i++;
 }
 j=GetTreeID(v->source);//choose the jth source tree
 t=MakeDNode(j,addno);//make the new beta
//Insert New Beta
 p=_opendnode->children;
 if(p==NULL)
  _opendnode->children=t;
 else
  {
   //insert in the increasing oder of address
   BOOL found=FALSE;
   q=p;
   while(p&&!found)
    if (p->addno>addno)
       found=TRUE;
    else
    {
     q=p;
     p=p->sibling;
    }
   if (!found)//insert at the end
    q->sibling=t;
   else
     if(p==q)//found at first child
      {
       _opendnode->children=t;
       t->sibling=p;
      }
      else //insert before p
       {
	q->sibling=t;
	t->sibling=p;
       }
   }
 return TRUE;
}
//*******************************************************
//This function add new beta tree in a full shape maner *
//*******************************************************
BOOL AddFullBeta(dnode *s)
{
 dnode *t,*p,*q;
 BYTE childrenno,addno;
 BOOL *check;
 address *v;
 snode *n;
 BYTE i,j;
//First count to see how many node is still open
_opendnodeno=0;
_countorfind=TRUE;//counting
_doitdnode=CountOpenDNode;
 PreorderDNode(s);//result is in _opendnodeno
if (_opendnodeno<=0)
 return FALSE;
//Then choose the first open node
_opendnodeno=1;
_opendnode=NULL;
_countorfind=FALSE;//finding
 LevelorderDNode(s);//result is _opendnode
//Then choose the first open address
 check=new BOOL [adtable[_opendnode->id].card]; assert(check!=NULL);
 for(i=0;i<adtable[_opendnode->id].card;i++)
  check[i]=FALSE;
 p=_opendnode->children;
 while(p)
 {
  check[p->addno]=TRUE;
  p=p->sibling;
 }
 childrenno=0;
 while (check[childrenno])
  childrenno++;
 delete [] check;
 addno=childrenno;//next address (in increasing maner)
 v= adtable[_opendnode->id].head;
 j=0; //find this address
 while(j<addno)
 {
  v=v->next;
  j++;
 }
 i=GetTreeID(v->source);//choose one beta tree as random
//Find it
 t=MakeDNode(i,addno);//creat the node
 p=_opendnode->children;
 if(p==NULL)
  _opendnode->children=t;
 else
  {
    //insert in the increasing oder of address
   BOOL found=FALSE;
   q=p;
   while(p&&!found)
    if (p->addno>addno)
       found=TRUE;
    else
    {
     q=p;
     p=p->sibling;
    }
   if (!found)//insert at the end
    q->sibling=t;
   else
     if(p==q)//found at first child
      {
       _opendnode->children=t;
       t->sibling=p;
      }
      else //insert before p
       {
	q->sibling=t;
	t->sibling=p;
       }
  }
 return TRUE;
}
//***************************************
//This Function delete randomly a leave *
//***************************************
int _leavecount;
void CountLeave(dnode *s)
{
if(s)
 {
  if (s->children==NULL)
  _leavecount++;
 }
}
void DeleteLeave(dnode *s,dnode *parent)
{
 dnode *t;
if (_leavecount>0)
 {
  if(s->children==NULL)
   {
    _leavecount--;
    if(_leavecount==0)
    {
     if(s==parent->children)
      parent->children=s->sibling;
     else
      {
       t=parent->children;
       while(t->sibling!=s)
	t=t->sibling;
       t->sibling=s->sibling;
      }
      DeleteSDNode(s);
     }
   }
   else
    {
     t=s->children;
     while(t)
      {
       DeleteLeave(t,s);
       t=t->sibling;
      }
    }
 }
}
BOOL DeleteBeta(dnode *s)
{
  if(s->children!=NULL)
  {
   _leavecount=0;
   _doitdnode=CountLeave;
   PreorderDNode(s);
   _leavecount=IRandom(1,_leavecount);
   DeleteLeave(s,NULL);
   return TRUE;
   }
  return FALSE;
}
//**************************************************************
//Pruning tree by deleting marked subtrees                     *
//(subtrees that rooted with negative id)(added inversion 3.1) *
//**************************************************************
void PruneDNode(dnode *t,dnode *parent)
{
 dnode *p,*q;
 if(t)
 {
  if(t->id<0) //if marked off
   {
    if(t->id>-numetree)//type1 sub-code
    {
     if(parent)//not pruning at root
     {
      p=parent->children;
      if(p==t) //if p is the first child
       parent->children=p->sibling;
      else
       {
        while(p->sibling!=t)
          p=p->sibling;
        p->sibling=t->sibling;
       } 
      DeleteDNode(t); 
     }
    }
   else//type2 sub-code
    {
      
     p=t->children;
     while(p->sibling)
      {
       q=p->sibling;
       DeleteDNode(p);
       p=q;
      }
    //delete the node
      p->addno=t->addno;
      q=parent->children;
     if(q==t)
       parent->children=p;
     else
      {
       while(q->sibling!=t) q=q->sibling;
       q->sibling=p;
      }
     p->sibling=t->sibling;   
     DeleteSDNode(t);
     PruneDNode(p,parent);
     } 
   }
  else
  {
   p=t->children;
   while(p)
    {
     q=p->sibling;
     PruneDNode(p,t);
     p=q;
    }
  }  
 }
}
//**************************************************************
//These fucntions are used to unmark the genes in one genotype *
//**************************************************************
void UnMarkingDNode(dnode *t)
{
 if (t->id<0)
  if(t->id<=-numetree)
   t->id=-(t->id+numetree);
  else 
  t->id=-t->id; 
}
void UnMarkingDtree(dnode *t)
{
 _doitdnode=UnMarkingDNode;
  PreorderDNode(t);
}
#endif
